(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright (c) 2013-2014 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: options.ml,v 1.37 2014/07/16 21:46:59 fclement Exp $ *)

module Nat = Unatural;;
module Cmd_opt = Ucommand_option;;

module Ncmd_opt = Unumber_command_option;;
module Scmd_opt = Ustring_command_option;;

let string_of_list s_of l =
  let len = Nat.to_int (Ulist.length l) in
  Ulist.fold_left_i
    (fun i accu x ->
      match i with
      | 0 -> accu ^ s_of x
      | n when n = len - 1 -> accu ^ ", and " ^ s_of x
      | _ -> accu ^ ", " ^ s_of x)
    "" l
;;

open Unumber_command_option_spec;;
open Refinement_types;;
open Options_types;;

let o_qual = Upervasives.qualify "Options";;

let o_failwith = Upervasives.failwith_qual o_qual;;
let o_invalid_arg = Upervasives.invalid_arg_qual o_qual;;

(* Algorithm configuration. *)

let { Uaref.get_list = get_option_external_program_command;
      add_item = add_option_external_program_command_item; _; } =
  Uaref.make_aref_list []
;;

let option_external_program_command =
  {
   Cmd_opt.key = "-external-program-command";
   spec = Arg.Rest add_option_external_program_command_item;
   doc =
     Printf.sprintf
       "<cmd> [<option>...] [<argument>...]\n\
        \tSet the path to the external program command to <cmd>.\n\
        \tAlso set optional options and arguments.\n\
        \tThis must be the last arguments of the present command."
  }
;;

module Dimension_structure =
struct
  type t = option_dimension;;
  
  let kind = Positive;;
  let name = "dimension";;
  let full_name = Printf.sprintf "%s of the parameter in each point" name;;
  let default = 1;;
  let opt_doc = None;;
end
;;

module Dimension = Ncmd_opt.Integer_Make (Dimension_structure);;
let option_dimension = Dimension.option
and get_option_dimension = Dimension.get
;;

module Dimension_management_structure =
struct
  type t = option_dimension_management;;
  
  let name = "dimension management";;
  let plural_name = Printf.sprintf "%ss" name;;
  
  let default = Vector
  and all = [ Vector; Best_component_only; ]
  ;;
  
  let of_string = function
    | "vector" | "v" -> Vector
    | "best_component_only" | "best_only" | "bco" -> Best_component_only
    | s ->
        o_failwith "dimension_management_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;
  
  let to_string = function
    | Vector -> "vector"
    | Best_component_only -> "best_component_only"
  ;;
end
;;

module Dimension_management = Scmd_opt.Make (Dimension_management_structure);;
let option_dimension_management = Dimension_management.option
and get_option_dimension_management = Dimension_management.get
;;

module Cutting_strategy_structure =
struct
  type t = option_cutting_strategy;;
  
  let name = "cutting strategy"
  and plural_name = "cutting strategies"
  ;;
  
  let default = Option_Overall_best
  and all = [ Option_Best_in_a_family; Option_Overall_best; ]
  ;;
  
  let of_string = function
    | "best_in_a_family" | "bf" | "family" | "f" -> Option_Best_in_a_family
    | "overall_best" | "ob" | "best" | "b" -> Option_Overall_best
    | s ->
        o_failwith "cutting_strategy_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;
  
  let to_string = function
    | Option_Best_in_a_family -> "best_in_a_family"
    | Option_Overall_best -> "overall_best"
  ;;
end
;;

module Cutting_strategy = Scmd_opt.Make (Cutting_strategy_structure);;
let option_cutting_strategy = Cutting_strategy.option
and get_option_cutting_strategy = Cutting_strategy.get
;;

module Number_of_cuts_structure =
struct
  type t = option_number_of_cuts;;
  
  let kind = Nonnegative;;
  let name = "number of cuts";;
  let full_name = Printf.sprintf "%s for each basic cutting family" name;;
  let default = 0;;
  let opt_doc =
    Some "By convention, 0 means all possible cuts of the family."
  ;;
end
;;

module Number_of_cuts = Ncmd_opt.Integer_Make (Number_of_cuts_structure);;
let option_number_of_cuts = Number_of_cuts.option
and get_option_number_of_cuts = Number_of_cuts.get
;;

let all_option_basic_cutting_families = [ Horizontal_cut; Vertical_cut; ];;

let default_option_basic_cutting_families = all_option_basic_cutting_families;;

let { Uaref.get_list = get_option_basic_cutting_families;
      clear_list = clear_option_basic_cutting_families;
      add_item = add_option_basic_cutting_family; } =
  Uaref.make_aref_list default_option_basic_cutting_families
;;

let option_clear_basic_cutting_families =
  {
   Cmd_opt.key = "-clear-basic-cutting-families";
   spec = Arg.Unit clear_option_basic_cutting_families;
   doc = Printf.sprintf "\n\tClear the basic cutting family list.";
  }
;;

let option_basic_cutting_family_of_string = function
  | "horizontal" | "hor" | "h" -> Horizontal_cut
  | "vertical" | "ver" | "v" -> Vertical_cut
  | s ->
      o_failwith "basic_cutting_family_of_string"
        (Printf.sprintf "Unknown basic cutting family %S" s)
;;

let add_option_basic_cutting_family_from_string =
  Upervasives.compose
    add_option_basic_cutting_family
    option_basic_cutting_family_of_string
;;

let string_of_option_basic_cutting_family = function
  | Horizontal_cut -> "horizontal"
  | Vertical_cut -> "vertical"
;;

let print_all_option_basic_cutting_families () =
  string_of_list
    string_of_option_basic_cutting_family
    all_option_basic_cutting_families
;;

let print_default_option_basic_cutting_family_list () =
  string_of_list
    string_of_option_basic_cutting_family
    (get_option_basic_cutting_families ())
;;

let option_add_basic_cutting_family =
  {
   Cmd_opt.key = "-add-basic-cutting-family";
   spec = Arg.String add_option_basic_cutting_family_from_string;
   doc =
     Printf.sprintf
       "<str>\n\
        \tAdd <str> to the list of basic cutting families.\n\
        \tUsed when the cutting strategy is set to %S.\n\
        \tValid basic cutting families are: %t.\n\
        \tDefault list is %t."
       (Cutting_strategy_structure.to_string Option_Best_in_a_family)
       print_all_option_basic_cutting_families
       print_default_option_basic_cutting_family_list;
  }
;;

module Candidate_cuttings_selection_criterion_ratio_structure =
struct
  type t = option_candidate_cuttings_selection_criterion_ratio;;
  
  let kind = Range ("0", "1");;
  let name = "ratio";;
  let full_name =
    Printf.sprintf "candidate cuttings selection criterion %s" name
  ;;
  let default = 0.8;;
  let opt_doc =
    Some "Candidate cuttings with indicator smaller than \
          the highest indicator\n\
          \ttimes <float> will not go through the optimization phase."
  ;;
end
;;

module Candidate_cuttings_selection_criterion_maximum_number_structure =
struct
  type t = option_candidate_cuttings_selection_criterion_maximum_number;;
  
  let kind = Positive;;
  let name = "maximum number";;
  let full_name =
    Printf.sprintf "candidate cuttings selection criterion %s" name
  ;;
  let default = 4;;
  let opt_doc =
    Some "At most <nat> candidate cuttings \
          will go through the optimization phase."
  ;;
end
;;

module Candidate_cuttings_selection_criterion_ratio =
  Ncmd_opt.Float_Make
    (Candidate_cuttings_selection_criterion_ratio_structure);;
let option_candidate_cuttings_selection_criterion_ratio =
  Candidate_cuttings_selection_criterion_ratio.option
and get_option_candidate_cuttings_selection_criterion_ratio =
  Candidate_cuttings_selection_criterion_ratio.get
;;

module Candidate_cuttings_selection_criterion_maximum_number =
  Ncmd_opt.Integer_Make
    (Candidate_cuttings_selection_criterion_maximum_number_structure);;
let option_candidate_cuttings_selection_criterion_maximum_number =
  Candidate_cuttings_selection_criterion_maximum_number.option
and get_option_candidate_cuttings_selection_criterion_maximum_number =
  Candidate_cuttings_selection_criterion_maximum_number.get
;;

module Minimum_gradient_structure =
struct
  type t = option_minimum_gradient;;
  
  let kind = Nonnegative;;
  let gname = "minimum gradient";;
  let name = gname;;
  let full_name = gname;;
  let default = 1e-4;;
  let opt_doc = None;;
end
;;

module Minimum_gradient = Ncmd_opt.Float_Make (Minimum_gradient_structure);;
let option_minimum_gradient = Minimum_gradient.option
and get_option_minimum_gradient = Minimum_gradient.get
;;

module Max_iteration_structure =
struct
  type t = option_max_iteration;;
  
  let kind = Nonnegative;;
  let name = "max iteration"
  and full_name = "maximum number of iterations"
  ;;
  let default = 10;;
  let opt_doc = None;;
end
;;

module Max_iteration = Ncmd_opt.Integer_Make (Max_iteration_structure);;
let option_max_iteration = Max_iteration.option
and get_option_max_iteration = Max_iteration.get
;;

module Min_best_scalar_indicator_structure =
struct
  type t = option_min_best_scalar_indicator;;
  
  let kind = Nonnegative;;
  let gname = "scalar indicator";;
  let name = Printf.sprintf "min best %s" gname;;
  let full_name = Printf.sprintf "minimum best %s" gname;;
  let default = 1e-2;;
  let opt_doc = None;;
end
;;

module Min_best_scalar_indicator =
  Ncmd_opt.Float_Make (Min_best_scalar_indicator_structure)
;;
let option_min_best_scalar_indicator = Min_best_scalar_indicator.option
and get_option_min_best_scalar_indicator = Min_best_scalar_indicator.get
;;

module Min_scalar_relative_error_structure =
struct
  type t = option_min_scalar_relative_error;;
  
  let kind = Nonnegative;;
  let gname = "relative error";;
  let name = Printf.sprintf "min scalar %s" gname;;
  let full_name = Printf.sprintf "minimum scalar %s" gname;;
  let default = 1e-2;;
  let opt_doc = None;;
end
;;

module Min_scalar_relative_error =
  Ncmd_opt.Float_Make (Min_scalar_relative_error_structure)
;;
let option_min_scalar_relative_error = Min_scalar_relative_error.option
and get_option_min_scalar_relative_error = Min_scalar_relative_error.get
;;

(* Program configuration. *)

module Out_formatter_structure =
struct
  type t = option_out_formatter;;
  
  let name = "out formatter";;
  let plural_name = Printf.sprintf "%ss" name;;
  
  let default = Option_Std_out
  and all =
    [ Option_Std_out; Option_Std_err; Option_Out_file "<file name>" ; ]
  ;;
  
  let of_string = function
    | "-" | "stdout" | "Std_out" -> Option_Std_out
    | "stderr" | "Std_err" -> Option_Std_err
    | "" ->
        o_invalid_arg "out_formatter_of_string"
          "Empty out formatter specification"
    | s -> Option_Out_file s
  ;;
  
  let to_string = function
    | Option_Std_out -> "Std_out"
    | Option_Std_err -> "Std_err"
    | Option_Out_file fname -> fname
  ;;
end
;;

module Out_formatter = Scmd_opt.Make (Out_formatter_structure);;
let option_out_formatter = Out_formatter.option
and get_option_out_formatter = Out_formatter.get
;;

let { Uaref.get_flag = get_option_report_allocated_memory;
      set_flag = set_option_report_allocated_memory; _; } =
  Uaref.make_flag ()
;;

let option_report_allocated_memory =
  {
   Cmd_opt.key = Ustring.long_key_of_string "report allocated memory";
   spec = Arg.Unit set_option_report_allocated_memory;
   doc =
     Printf.sprintf "\n\tReport total allocated memory. Default is no report.";
  }
;;

let { Uaref.get_flag = get_option_incremental_mode;
      set_flag = set_option_incremental_mode; _; } =
  Uaref.make_flag ()
;;

let option_incremental_mode =
  {
   Cmd_opt.key = Ustring.long_key_of_string "incremental mode";
   spec = Arg.Unit set_option_incremental_mode;
   doc =
     Printf.sprintf "\n\tTurn the incremental mode on. Default is off.";
  }
;;

module Graphics_mode_structure =
struct
  type t = option_graphics_mode;;
  
  let name = "graphics mode";;
  let plural_name = Printf.sprintf "%ss" name;;
  
  let default = Full_graphics
  and all = [ No_graphics; Parameter_graphics; Full_graphics; ]
  ;;
  
  let of_string = function
    | "no_graphics" | "ng" -> No_graphics
    | "parameter_graphics" | "pg" -> Parameter_graphics
    | "full_graphics" | "fg" -> Full_graphics
    | s ->
        o_failwith "graphics_mode_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;
  
  let to_string = function
    | No_graphics -> "no_graphics"
    | Parameter_graphics -> "parameter_graphics"
    | Full_graphics -> "full_graphics"
  ;;
end
;;

module Graphics_mode = Scmd_opt.Make (Graphics_mode_structure);;
let option_graphics_mode = Graphics_mode.option
and get_option_graphics_mode = Graphics_mode.get
;;

module Graphics_window_min_width_structure =
struct
  type t = option_graphics_window_min_width;;
  
  let kind = Positive;;
  let gname = "width";;
  let name = Printf.sprintf "graphics window min %s" gname;;
  let full_name = Printf.sprintf "graphics window minimum %s" gname;;
  let default = 800;;
  let opt_doc = None;;
end
;;

module Graphics_window_max_width_structure =
struct
  type t = option_graphics_window_max_width;;
  
  let kind = Positive;;
  let gname = "width";;
  let name = Printf.sprintf "graphics window max %s" gname;;
  let full_name = Printf.sprintf "graphics window maximum %s" gname;;
  let default = 1920;;
  let opt_doc = None;;
end
;;

module Graphics_window_min_height_structure =
struct
  type t = option_graphics_window_min_height;;
  
  let kind = Positive;;
  let gname = "height";;
  let name = Printf.sprintf "graphics window min %s" gname;;
  let full_name = Printf.sprintf "graphics window minimum %s" gname;;
  let default = 600;;
  let opt_doc = None;;
end
;;

module Graphics_window_max_height_structure =
struct
  type t = option_graphics_window_max_height;;
  
  let kind = Positive;;
  let gname = "height";;
  let name = Printf.sprintf "graphics window max %s" gname;;
  let full_name = Printf.sprintf "graphics window maximum %s" gname;;
  let default = 1080;;
  let opt_doc = None;;
end
;;

module Graphics_window_min_width =
  Ncmd_opt.Integer_Make (Graphics_window_min_width_structure)
;;
module Graphics_window_max_width =
  Ncmd_opt.Integer_Make (Graphics_window_max_width_structure)
;;
module Graphics_window_min_height =
  Ncmd_opt.Integer_Make (Graphics_window_min_height_structure)
;;
module Graphics_window_max_height =
  Ncmd_opt.Integer_Make (Graphics_window_max_height_structure)
;;

let option_graphics_window_min_width =
  Graphics_window_min_width.option
and option_graphics_window_max_width =
  Graphics_window_max_width.option
and option_graphics_window_min_height =
  Graphics_window_min_height.option
and option_graphics_window_max_height =
  Graphics_window_max_height.option
;;

let get_option_graphics_window_min_width =
  Graphics_window_min_width.get
and get_option_graphics_window_max_width =
  Graphics_window_max_width.get
and get_option_graphics_window_min_height =
  Graphics_window_min_height.get
and get_option_graphics_window_max_height =
  Graphics_window_max_height.get
;;

let color_scale_name_of_string name = function
  | "gray" | "grey" | "g" -> Gray
  | "cool" | "c" -> Cool
  | "hot" | "h" -> Hot
  | "jet" | "j" -> Jet
  | s ->
      o_failwith "color_scale_name_of_string"
        (Printf.sprintf "Unknown %s %S" name s)
;;

let color_scale_name_to_string = function
  | Gray -> "gray"
  | Cool -> "cool"
  | Hot -> "hot"
  | Jet -> "jet"
;;

let all_color_scale_names = [ Gray; Cool; Hot; Jet; ];;

module Graphics_parameter_color_scale_name_structure =
struct
  type t = option_graphics_color_scale_name;;
  
  let name = "graphics parameter color scale name";;
  let plural_name = Printf.sprintf "%ss" name;;
  
  let default = Hot
  and all = all_color_scale_names
  and of_string = color_scale_name_of_string name
  and to_string = color_scale_name_to_string
  ;;
end
;;

module Graphics_parameter_color_scale_name =
  Scmd_opt.Make (Graphics_parameter_color_scale_name_structure)
;;
let option_graphics_parameter_color_scale_name =
  Graphics_parameter_color_scale_name.option
and get_option_graphics_parameter_color_scale_name =
  Graphics_parameter_color_scale_name.get
;;

module Graphics_parameter_min_structure =
struct
  type t = option_graphics_parameter_min;;
  
  let kind = Nonnegative;;
  let gname = "parameter";;
  let name = Printf.sprintf "graphics minimum %s" gname;;
  let full_name = name;;
  let default = 0.;;
  let opt_doc =
    Some "Parameters under this value are clipped when displayed."
  ;;
end
;;

module Graphics_parameter_max_structure =
struct
  type t = option_graphics_parameter_max;;
  
  let kind = Nonnegative;;
  let gname = "parameter";;
  let name = Printf.sprintf "graphics maximum %s" gname;;
  let full_name = name;;
  let default = 1.;;
  let opt_doc =
    Some "Parameters over this value are clipped when displayed."
  ;;
end
;;

module Graphics_parameter_min =
  Ncmd_opt.Float_Make (Graphics_parameter_min_structure)
;;
module Graphics_parameter_max =
  Ncmd_opt.Float_Make (Graphics_parameter_max_structure)
;;

let option_graphics_parameter_min = Graphics_parameter_min.option
and option_graphics_parameter_max = Graphics_parameter_max.option
;;

let get_option_graphics_parameter_min = Graphics_parameter_min.get
and get_option_graphics_parameter_max = Graphics_parameter_max.get
;;

module Graphics_gradient_color_scale_name_structure =
struct
  type t = option_graphics_color_scale_name;;
  
  let name = "graphics gradient color scale name";;
  let plural_name = Printf.sprintf "%ss" name;;
  
  let default = Gray
  and all = all_color_scale_names
  and of_string = color_scale_name_of_string name
  and to_string = color_scale_name_to_string
  ;;
end
;;

module Graphics_gradient_color_scale_name =
  Scmd_opt.Make (Graphics_gradient_color_scale_name_structure)
;;
let option_graphics_gradient_color_scale_name =
  Graphics_gradient_color_scale_name.option
and get_option_graphics_gradient_color_scale_name =
  Graphics_gradient_color_scale_name.get
;;

module Graphics_gradient_amplitude_structure =
struct
  type t = option_graphics_gradient_amplitude;;
  
  let kind = Positive;;
  let gname = "gradient amplitude";;
  let name = Printf.sprintf "graphics %s" gname;;
  let full_name = name;;
  let default = 1.;;
  let opt_doc =
    Some
      "Gradients with absolute value over this value, are clipped\n\
       \twhen displayed."
  ;;
end
;;

module Graphics_gradient_amplitude =
  Ncmd_opt.Float_Make (Graphics_gradient_amplitude_structure)
;;
let option_graphics_gradient_amplitude = Graphics_gradient_amplitude.option
and get_option_graphics_gradient_amplitude = Graphics_gradient_amplitude.get
;;

(* Command-line options. *)

let regular_opt rco =
  Cmd_opt.with_auto_short_key
    (Cmd_opt.with_auto_gnu_key
       (Cmd_opt.of_regular rco))
and regular_opt_with_short_key (sk, rco) =
  Cmd_opt.with_short_key sk
    (Cmd_opt.with_auto_gnu_key
       (Cmd_opt.of_regular rco))
;;

let opt = function
  | (None, rco) -> regular_opt rco
  | (Some sk, rco) -> regular_opt_with_short_key (sk, rco)
;;

let user_add_all l = Cmd_opt.user_add_all (Ulist.map opt l);;

let phase_dimension = Cmd_opt.Phase "dimension";;

let add_options () =
  user_add_all
    [
     (Some "--", option_external_program_command);
     (Some "-dim", option_dimension);
     (None, option_dimension_management);
     (None, option_cutting_strategy);
     (Some "-nc", option_number_of_cuts);
     (Some "-clear-bcf", option_clear_basic_cutting_families);
     (Some "-add-bcf", option_add_basic_cutting_family);
     (Some "-sel-ratio", option_candidate_cuttings_selection_criterion_ratio);
     (Some "-sel-num",
      option_candidate_cuttings_selection_criterion_maximum_number);
     (Some "-min-g", option_minimum_gradient);
     (Some "-stop-it", option_max_iteration);
     (Some "-stop-ind", option_min_best_scalar_indicator);
     (Some "-stop-cost", option_min_scalar_relative_error);
     (None, option_out_formatter);
     (None, option_report_allocated_memory);
     (None, option_incremental_mode);
     (None, option_graphics_mode);
     (Some "-gr-min-w", option_graphics_window_min_width);
     (Some "-gr-max-w", option_graphics_window_max_width);
     (Some "-gr-min-h", option_graphics_window_min_height);
     (Some "-gr-max-h", option_graphics_window_max_height);
     (Some "-gr-cs-p", option_graphics_parameter_color_scale_name);
     (Some "-gr-min-p", option_graphics_parameter_min);
     (Some "-gr-max-p", option_graphics_parameter_max);
     (Some "-gr-cs-g", option_graphics_gradient_color_scale_name);
     (Some "-gr-amp-g", option_graphics_gradient_amplitude);
    ];
  Cmd_opt.phase_add phase_dimension "-dimension";
;;

add_options ();;

Cmd_opt.set_usage_msg
  (Format.sprintf
     "@.%s -- Adaptive parameterization@.@.\
      Usage:@.@.\
      %s [<option>...] -- <cmd> [<cmd_option>...] [<cmd_argument>...]@.@.\
      Options are:@."
     Sys.argv.(0) Sys.argv.(0))
;;

(* Report/debug flags and warning/error messages. *)

let purpose = Udebug.User;;
let software_name = Ref_indic_version.pph_id_card.Upph_id_card.project_name;;
let key_prefixes = Udebug.make_key_prefixes software_name;;

let printing_facilities =
  Udebug.make_printing_facilities purpose software_name key_prefixes
;;
